"""Module containing image transformation functions.

This module contains the function decorator
:func:`jicbioimage.core.transform.transformation` that can be used
to turn functions into image transformations.

Below is an example of how to create a transformation that inverts an image.

>>> import numpy as np
>>> @transformation
... def invert(image):
...     "Return the inverted image."
...     maximum = np.iinfo(image.dtype).max
...     maximum_array = np.ones(image.shape, dtype=image.dtype) * maximum
...     return maximum_array - image
...

"""

from functools import wraps

from jicbioimage.core.io import AutoName, AutoWrite
from jicbioimage.core.image import Image


def transformation(func):
    """Function decorator to turn another function into a transformation."""
    @wraps(func)
    def func_as_transformation(*args, **kwargs):

        # Get the input image, so that we can get the history from it.
        input_image = kwargs.get("image", None)
        if input_image is None:
            input_image = args[0]

        # Get the history from the image.
        history = []
        if hasattr(input_image, "history"):
            history.extend(input_image.history)

        image = func(*args, **kwargs)
        if not isinstance(image, Image):
            image = Image.from_array(image, log_in_history=False)

        # Update the history of the image.
        image.history = history
        image.history.append('Applied {} transform'.format(func.__name__))

        if AutoWrite.on:
            fpath = AutoName.name(func)
            with open(fpath, "wb") as fh:
                fh.write(image.png())
        return image
    return func_as_transformation



"""Module containing utility functions for manipulating numpy arrays."""

import sys
from functools import wraps

import random

import numpy as np

from jicbioimage.core.util.color import pretty_color_palette

def normalise(array):
    """Return array normalised such that all values are between 0 and 1.

    If all the values in the array are the same the function will return:
    - np.zeros(array.shape, dtype=np.float) if the value is 0 or less
    - np.ones(array.shape, dtype=np.float) if the value is greater than 0

    :param array: numpy.array
    :returns: numpy.array.astype(numpy.float)
    """
    min_val = array.min()
    max_val = array.max()
    array_range = max_val - min_val

    if array_range == 0:
        # min_val == max_val
        if min_val > 0:
            return np.ones(array.shape, dtype=np.float)
        return np.zeros(array.shape, dtype=np.float)

    return (array.astype(np.float) - min_val) / array_range


def reduce_stack(array3D, z_function):
    """Return 2D array projection of the input 3D array.

    The input function is applied to each line of an input x, y value.

    :param array3D: 3D numpy.array
    :param z_function: function to use for the projection (e.g. :func:`max`)
    """
    xmax, ymax, _ = array3D.shape
    projection = np.zeros((xmax, ymax), dtype=array3D.dtype)
    for x in range(xmax):
        for y in range(ymax):
            projection[x, y] = z_function(array3D[x, y, :])
    return projection


def map_stack(array3D, z_function):
    """Return 3D array where each z-slice has had the function applied to it.

    :param array3D: 3D numpy.array
    :param z_function: function to be mapped to each z-slice
    """
    _, _, zdim = array3D.shape
    return np.dstack([z_function(array3D[:, :, z]) for z in range(zdim)])


def check_dtype(array, allowed):
    """Raises TypeError if the array is not of an allowed dtype.

    :param array: array whose dtype is to be checked
    :param allowed: instance or list of allowed dtypes
    """
    if not hasattr(allowed, "__iter__"):
        allowed = [allowed, ]
    if array.dtype not in allowed:
        msg = "Invalid dtype {}. Allowed dtype(s): {}"
        raise(TypeError(msg.format(array.dtype, allowed)))


def dtype_contract(input_dtype=None, output_dtype=None):
    """Function decorator for specifying input and/or output array dtypes.

    :param input_dtype: dtype of input array
    :param output_dtype: dtype of output array
    :returns: function decorator
    """
    def wrap(function):
        @wraps(function)
        def wrapped_function(*args, **kwargs):
            if input_dtype is not None:
                check_dtype(args[0], input_dtype)
            array = function(*args, **kwargs)
            if output_dtype is not None:
                check_dtype(array, output_dtype)
            return array
        return wrapped_function
    return wrap


def false_color(array, color_dict=None, keep_zero_black=True):
    """Return a RGB false color array.

    Assigning a unique RGB color value to each unique element of the input
    array and return an array of shape (array.shape, 3).

    :param array: input numpy.array
    :param color_dict: dictionary with keys/values corresponding to identifiers
                       and RGB tuples respectively
    :param keep_zero_black: whether or not the background should be black
    :returns: numpy.array
    """

    output_array = np.zeros(array.shape + (3,), np.uint8)

    unique_identifiers = set(np.unique(array))

    if color_dict is None:
        color_dict = pretty_color_palette(
            unique_identifiers, keep_zero_black)

    for identifier in unique_identifiers:
        output_array[np.where(array == identifier)] = color_dict[identifier]

    return output_array



"""Module for generating RGB tuples for use as colors in images."""

import sys
import random
import hashlib
import itertools

def _extract_8_bits(long_value, shift=1):
    """Return an integer in the range 0, 255 by extracting 8 bits from the
    input long value. shift determines which 8 bits are taken, by default
    the first 8 bits."""

    bitmask = (1 << 8 * shift) - 1

    return (long_value & bitmask) >> (8 * (shift-1))

def _md5_hash_as_long(input_value):
    """Return the hash of the input value converted to a long."""

    hex_hash = hashlib.md5(str(input_value).encode('utf-8')).hexdigest()

    return int(hex_hash, 16)


def _generate_rgb_tuple(long_hash):
    """Return an RGB tuple with the following ranges:

    0 <= blue < 128
    128 <= green < 256
    0 <= red < 256"""

    blue = _extract_8_bits(long_hash, 1) & 127
    green = 128 + (_extract_8_bits(long_hash, 2) & 127)
    red = _extract_8_bits(long_hash, 3)

    return map(int, (red, green, blue))

def pretty_color(identifier=None):
    """Return aesthetically pleasing RGB tuple.

    :returns: RGB tuple
    """

    if identifier is None:
        try:
            identifier = random.randint(0, sys.maxint)
        except AttributeError:
            # Python3 has no sys.maxint
            identifier = random.randint(0, sys.maxsize)

    long_hash = _md5_hash_as_long(identifier)

    tuple_as_integers = _generate_rgb_tuple(long_hash)

    tuple_permutations = list(itertools.permutations(tuple_as_integers))

    selected_permutation = _extract_8_bits(long_hash, 4) % 6

    return tuple_permutations[selected_permutation]

def pretty_color_palette(identifiers, keep_zero_black=True):
    """Return dictionary with pretty colors.

    :param identifiers: set of unique identifiers
    :param keep_zero_black: whether or not the background should be black
    :returns: dictionary
    """

    color_dict = {}
    for i in identifiers:
        if keep_zero_black and i == 0:
            color_dict[0] = (0, 0, 0)
            continue
        value = pretty_color(i)
        color_dict[i] = value

    return color_dict



"""Module for reading and writing images."""

import sys
import os.path
import re
import subprocess
import json
from collections import namedtuple
import hashlib

from jicbioimage.core.image import ImageCollection, MicroscopyCollection


def _md5_hexdigest_from_file(fpath, blocksize=65536):
    """Return md5 hex digest of a file."""
    md5_hash = hashlib.md5()
    with open(fpath, "rb") as fh:
        buf = fh.read(blocksize)
        while len(buf) > 0:
            md5_hash.update(buf)
            buf = fh.read(blocksize)
        return md5_hash.hexdigest()


class AutoName(object):
    """Class for generating output file names automatically."""
    count = 0
    directory = None  #: Output directory to save images to.
    suffix = ".png"   #: Image file suffix.
    prefix_format = "{:d}_"  #: Image file prefix format.
    namespace = ""  #: Image file namespace.

    @classmethod
    def prefix(cls):
        """Return auto generated file prefix."""
        return cls.prefix_format.format(cls.count)

    @classmethod
    def name(cls, func):
        """Return auto generated file name."""
        cls.count = cls.count + 1
        fpath = '{}{}{}{}'.format(cls.prefix(), cls.namespace,
                                  func.__name__, cls.suffix)
        if cls.directory:
            fpath = os.path.join(cls.directory, fpath)
        return fpath


class AutoWrite(object):
    """Class for writing images automatically."""

    #: Whether or not auto writing of images is enabled.
    on = True


#############################################################################
# Back ends classes for storing/caching unpacked microscopy images.
#############################################################################

class FileBackend(object):
    """Class for storing image files."""

    class Entry(object):
        """Class representing a backend entry."""
        def __init__(self, base_dir, fpath):
            """Initialise a new entry; to be populated with images.

            The base name of the fpath argument is used to create a
            subdirectory in the backend directory specific for the microscopy
            file to be loaded.

            :param base_dir: backend directory
            :param fpath: path to the microscopy image of interest
            """
            md5 = _md5_hexdigest_from_file(fpath)
            self._directory = os.path.join(base_dir, md5)
            if not os.path.isdir(self.directory):
                os.mkdir(self.directory)

        @property
        def directory(self):
            """Where the images are stored."""
            return self._directory

    def __init__(self, directory):
        """Initialise a backend.

        Creates the backend directory if it does not already exist.

        :param directory: location of the backend
        """
        if not os.path.isdir(directory):
            os.mkdir(directory)
        self._directory = directory

    @property
    def directory(self):
        """Where the entries are stored."""
        return self._directory

    def new_entry(self, fpath):
        """Return a new entry; to be populated with images.

        :param fpath: path to microscopy image
        :returns: :class:`jiciimagelib.image.FileBackend.Entry` instance
        """
        return FileBackend.Entry(self.directory, fpath)


class BFConvertWrapper(object):
    """Class for unpacking microscopy files using bfconvert."""

    def __init__(self, backend):
        self.backend = backend
        self.split_order = ['s', 'c', 'z', 't']

    def split_pattern(self, win32=False):
        """Pattern used to split the input file."""
        patterns = []
        for p in self.split_order:
            if win32:
                patterns.append('{}%%{}'.format(p.capitalize(), p))
            else:
                patterns.append('{}%{}'.format(p.capitalize(), p))
        return '_'.join(patterns)

    def _sorted_nicely(self, l):
        """Return list sorted in the way that humans expect.

        :param l: iterable to be sorted
        :returns: sorted list
        """
        convert = lambda text: int(text) if text.isdigit() else text
        sort_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
        return sorted(l, key=sort_key)

    def manifest(self, entry):
        """Returns manifest as a list.

        :param entry: :class:`jicbioimage.core.image.FileBackend.Entry`
        :returns: list
        """
        entries = []
        for fname in self._sorted_nicely(os.listdir(entry.directory)):
            if fname == 'manifest.json':
                continue
            fpath = os.path.abspath(os.path.join(entry.directory, fname))
            metadata = self.metadata_from_fname(fname)
            entries.append({"filename": fpath,
                            "series": metadata.s,
                            "channel": metadata.c,
                            "zslice": metadata.z,
                            "timepoint": metadata.t})
        return entries

    def run_command(self, input_file, output_dir=None):
        """Return the command for running bfconvert as a list.

        :param input_file: path to microscopy image to be converted
        :param ouput_dir: directory to write output tiff files to
        :returns: list
        """
        base_name = os.path.basename(input_file)
        name, suffix = base_name.split('.', 1)
        output_file = '{}.tif'.format(self.split_pattern())
        bfconvert = 'bfconvert'
        if sys.platform == 'win32':
            bfconvert = 'bfconvert.bat'
            output_file = '{}.tif'.format(self.split_pattern(win32=True))
        if output_dir:
            output_file = os.path.join(output_dir, output_file)
        return [bfconvert, input_file, output_file]

    def metadata_from_fname(self, fname):
        """Return meta data extracted from file name.

        :param fname: metadata file name
        :returns: dynamically created :class:`collections.namedtuple`
        """
        MetaData = namedtuple('MetaData', self.split_order)

        base_name = os.path.basename(fname)
        # e.g. 'S1_C2_Z3_T4.tif'

        name, suffix = base_name.split('.')
        # e.g. 'S1_C2_Z3_T4', 'tif'

        data = name.split('_')
        # e.g. ['S1', 'C2', 'Z3', 'T4']

        args = [int(x[1:]) for x in data]
        # e.g. [1, 2, 3, 4]

        return MetaData(*args)

    def already_converted(self, fpath):
        """Return true if the file already has a manifest file in the backend.

        :param fpath: potential path to the manifest file
        :returns: bool
        """
        manifest_fpath = os.path.join(self.backend.directory,
                                      _md5_hexdigest_from_file(fpath),
                                      'manifest.json')
        return os.path.isfile(manifest_fpath)

    def __call__(self, input_file):
        """Run the conversion.

        Unpacks the microscopy file and creates the manifest file.

        :param input_file: path to the microscopy file
        :returns: path to manifest file
        """
        entry = self.backend.new_entry(input_file)
        cmd = self.run_command(input_file, entry.directory)
        try:
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            stderr = p.stderr.read()
        except OSError as e:
            msg = 'bfconvert tool not found in PATH\n{}'.format(e)
            raise(RuntimeError(msg))
        if len(stderr) > 0:
            raise(RuntimeError(stderr))
        manifest_fpath = os.path.join(entry.directory, 'manifest.json')
        with open(manifest_fpath, 'w') as fh:
            json.dump(self.manifest(entry), fh)
        return manifest_fpath


class DataManager(list):
    """Manage :class:`jicbioimage.core.image.ImageCollection` instances."""

    def __init__(self, backend=None):
        if backend is None:
            dirpath = os.path.join(os.getcwd(), 'jicbioimage.core_backend')
            backend = FileBackend(directory=dirpath)
        self.backend = backend
        self.convert = BFConvertWrapper(self.backend)

    def load(self, fpath):
        """Load a microscopy file.

        :param fpath: path to microscopy file
        """
        def is_microscopy_item(fpath):
            """Return True if the fpath is likely to be microscopy data.

            :param fpath: file path to image
            :returns: :class:`bool`
            """
            l = fpath.split('.')
            ext = l[-1]
            pre_ext = l[-2]
            if ((ext == 'tif' or ext == 'tiff')
               and pre_ext != 'ome'):
                return False
            return True

        if not self.convert.already_converted(fpath):
            path_to_manifest = self.convert(fpath)
        else:
            path_to_manifest = os.path.join(self.backend.directory,
                                            _md5_hexdigest_from_file(fpath),
                                            'manifest.json')

        collection = None
        if is_microscopy_item(fpath):
            collection = MicroscopyCollection(path_to_manifest)
        else:
            collection = ImageCollection(path_to_manifest)
        self.append(collection)

        return collection
